"""Kukur makes time series data and metadata available to the Apache Arrow ecosystem."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
import typing

from datetime import datetime
from typing import Generator, Optional, Protocol, Union

import pyarrow as pa

from .base import (
    DataType,
    Dictionary,
    InterpolationType,
    SeriesSearch,
    SeriesSelector,
    SourceStructure,
)
from .exceptions import KukurException  # noqa
from .metadata import Metadata
from .client import Client


@typing.runtime_checkable
class Source(Protocol):
    """Source is the interface that Kukur data sources need to implement."""

    def search(
        self, selector: SeriesSearch
    ) -> Generator[Union[SeriesSelector, Metadata], None, None]:
        """Return all time series or even the metadata of them in this source matching the selector."""
        ...

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Return metadata for the given time series."""
        ...

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Return data for the given time series in the given time period."""
        ...


@typing.runtime_checkable
class TagSource(Source, Protocol):
    """TagSource is the interface that Kukur data sources that support tags and fields need to implement."""

    def get_source_structure(
        self, selector: SeriesSelector
    ) -> Optional[SourceStructure]:
        """Return the available tag keys and tag values and fields of a source."""
        ...


@typing.runtime_checkable
class PlotSource(Source, Protocol):
    """PlotSource is the interface that Kukur data sources that support getting plot data need to implement."""

    def get_plot_data(
        self,
        selector: SeriesSelector,
        start_date: datetime,
        end_date: datetime,
        interval_count: int,
    ) -> pa.Table:
        """Return plotting data for the given time series in the given time period."""
        ...


@typing.runtime_checkable
class SignalGenerator(Protocol):
    """Protocol for generating signals."""

    def generate(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Generates data based on a selector, start date and end date."""
        ...

    def list_series(self, selector: SeriesSearch) -> Generator[Metadata, None, None]:
        """Yields all possible metadata combinations using the signal configuration and the provided selector."""
        ...


__all__ = [
    "Client",
    "DataType",
    "Dictionary",
    "InterpolationType",
    "Metadata",
    "SeriesSearch",
    "SeriesSelector",
]



"""Common parts for all repositories."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
import abc
import contextlib
import sqlite3

from typing import List
from dateutil.parser import parse as parse_date


class BaseRepository:  # pylint: disable=too-few-public-methods
    """Base repository to open a fully-featured connection to sqlite."""

    __connection_string: str

    def __init__(self, connection_string: str):
        self.__connection_string = connection_string

    def _get_connection_string(self) -> str:
        return self.__connection_string

    def _open_db(self):
        return _open_db(self._get_connection_string())


class Migration(abc.ABC):  # pylint: disable=too-few-public-methods
    """A database migration makes changes to the schema of the database."""

    __connection_string: str

    def __init__(self, connection_string: str):
        self.__connection_string = connection_string

    @abc.abstractmethod
    def migrate(self):
        """Run a migration in the database given by the cursor."""
        ...

    def _open_db(self):
        return _open_db(self.__connection_string)


class MigrationRunner:
    """MigrationRunner keeps track of registered database migrations and runs them."""

    __migrations: List[Migration]

    def __init__(self):
        self.__migrations = []

    def register(self, migration: Migration):
        """Register a new migration with the migration runner."""
        self.__migrations.append(migration)

    def migrate(self):
        """Run all registered migrations."""
        for migration in self.__migrations:
            migration.migrate()


def _open_db(connection_string: str):
    sqlite3.register_converter("datetime", parse_date)
    connection = sqlite3.connect(
        connection_string,
        uri=True,
        timeout=60.0,
        detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
    )
    connection.execute("pragma foreign_keys = ON")
    return contextlib.closing(connection)



"""Kukur connection to SQLite."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

import re
import sqlite3

from dateutil.parser import parse as parse_date

from kukur.source.metadata import MetadataValueMapper
from kukur.source.quality import QualityMapper
from kukur.source.sql import BaseSQLSource, SQLConfig


sqlite3.register_converter("datetime", parse_date)


def _match(pattern, value):
    match = re.search(pattern, value)
    if match is None:
        return None
    return match.group(1)


class SQLiteSource(BaseSQLSource):
    """Kukur source for SQLite."""

    def connect(self):
        uri = False
        if self._config.connection_string.startswith("file:"):
            uri = True
        if self._config.query_timeout_seconds == 0:
            connection = sqlite3.connect(
                self._config.connection_string,
                detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
                uri=uri,
            )
        else:
            connection = sqlite3.connect(
                self._config.connection_string,
                timeout=self._config.query_timeout_seconds,
                detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
                uri=uri,
            )

        connection.create_function("match", 2, _match)
        return connection


def from_config(
    data, metadata_value_mapper: MetadataValueMapper, quality_mapper: QualityMapper
) -> SQLiteSource:
    """Create a new SQLite source from a configuration dictionay."""

    config = SQLConfig.from_dict(data)
    return SQLiteSource(config, metadata_value_mapper, quality_mapper)



"""Return fake data for integration tests."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from typing import Generator, Union

from pyarrow import Table

from kukur import Metadata, SeriesSearch, SeriesSelector
from kukur.exceptions import InvalidDataError
from kukur.metadata.fields import Description, Unit


class IntegrationTestSource:
    """Source that returns fake data for Flight integration testing."""

    def search(  # pylint: disable=no-self-use
        self, selector: SeriesSearch
    ) -> Generator[Union[Metadata, SeriesSelector], None, None]:
        """Search time series using the Flight service."""
        yield SeriesSelector(
            selector.source, {"tag1": "value1", "tag2": "value2"}, "pressure"
        )
        yield Metadata(
            SeriesSelector(
                selector.source, {"tag1": "value1a", "tag2": "value2a"}, "temperature"
            ),
            {Description: "integration test temperature"},
        )
        yield Metadata(
            SeriesSelector(
                selector.source, {"tag1": "value1b", "tag2": "value2b"}, "pH"
            ),
            {Description: "integration test pH"},
        )

    def get_metadata(  # pylint: disable=no-self-use
        self, selector: SeriesSelector
    ) -> Metadata:
        """Get metadata from the Flight service."""
        if selector == SeriesSelector(
            selector.source, {"tag1": "value1", "tag2": "value2"}, "pressure"
        ):
            return Metadata(selector, {Description: "integration test pressure"})
        if selector == SeriesSelector(
            selector.source, {"tag1": "value1a", "tag2": "value2a"}, "temperature"
        ):
            return Metadata(selector, {Unit: "c"})
        if selector == SeriesSelector(
            selector.source, {"tag1": "value1b", "tag2": "value2b"}, "pH"
        ):
            raise InvalidDataError("Metadata not found")
        return Metadata(selector)

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> Table:
        """Get data from the Flight service."""


def from_config() -> IntegrationTestSource:
    """Create a new integration test source."""
    return IntegrationTestSource()



"""Metadata mappings for Kukur data sources."""

from typing import Dict, List, Union


class MetadataMapper:
    """MetadataMapper maps names for metadata fields used in a source to the names known by Kukur."""

    __mapping: Dict[str, str]

    @classmethod
    def from_config(cls, config: Dict[str, str]) -> "MetadataMapper":
        """Create a new mapper from a dictionary that maps Kukur names to external names."""
        mapper = cls()
        for k, v in config.items():
            mapper.add_mapping(k, v)
        return mapper

    def __init__(self):
        self.__mapping = {}

    def add_mapping(self, kukur_field_name: str, external_field_name: str):
        """Add a mapping"""
        self.__mapping[kukur_field_name] = external_field_name

    def from_kukur(self, kukur_field_name: str) -> str:
        """Map a metadata field name defined by kukur to an external field name.

        Returns the field name as is when no mapping is defined."""
        return self.__mapping.get(kukur_field_name, kukur_field_name)

    def from_source(self, external_field_name: str) -> str:
        """Return the field name mapping from source.

        Returns the field name as is when no mapping is defined."""
        reverse_mapping = {v: k for k, v in self.__mapping.items()}
        return reverse_mapping.get(external_field_name, external_field_name)


class MetadataValueMapper:
    """MetadataFieldMapper maps values for a metadata field to values known by Kukur."""

    __mapping: Dict[str, Dict[Union[str, int], str]]

    def __init__(self):
        self.__mapping = {}

    @classmethod
    def from_config(
        cls, config: Dict[str, Dict[str, Union[str, List[str]]]]
    ) -> "MetadataValueMapper":
        """Create a new mapper from a double dictionary

        The double dictionary maps kukur field names and kukur field values to external values."""
        mapper = cls()
        for field_name, field_mapping in config.items():
            for field_value, external_field_value in field_mapping.items():
                if isinstance(external_field_value, (str, int)):
                    mapper.add_mapping(field_name, field_value, external_field_value)
                else:
                    for choice in external_field_value:
                        mapper.add_mapping(field_name, field_value, choice)
        return mapper

    def add_mapping(
        self,
        field_name: str,
        kukur_field_value: str,
        external_field_value: Union[str, int],
    ):
        """Add a mapping for a metadata field value known to kukur from the external value."""
        if field_name not in self.__mapping:
            self.__mapping[field_name] = {}
        self.__mapping[field_name][external_field_value] = kukur_field_value

    def from_source(
        self, field_name: str, external_field_value: Union[str, int]
    ) -> str:
        """Map a field value as it's known in an external source to one known by Kukur."""
        if field_name not in self.__mapping:
            return str(external_field_value)
        return self.__mapping[field_name].get(
            external_field_value, str(external_field_value)
        )



"""Connect to another Kukur instance using Arrow Flight."""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from typing import Any, Dict, Generator, Optional, Tuple, Union

import pyarrow as pa

from kukur import Metadata, SeriesSearch, SeriesSelector, SourceStructure
from kukur.client import Client


def from_config(config: Dict[str, Any]):
    """Create a new Kukur source"""
    source_name = config["source"]
    host = config.get("host", "localhost")
    port = config.get("port", 8081)
    api_key = (config.get("api_key_name", ""), config.get("api_key", ""))
    return KukurSource(source_name, host, port, api_key)


class KukurSource:
    """KukurSource connects to another Kukur instance using Arrow Flight."""

    __client: Client
    __source_name: str

    # pylint: disable=too-many-arguments
    def __init__(
        self, source_name: str, host: str, port: int, api_key: Tuple[str, str]
    ):
        self.__client = Client(api_key, host, port)
        self.__source_name = source_name

    def search(
        self, selector: SeriesSearch
    ) -> Generator[Union[Metadata, SeriesSelector], None, None]:
        """Search time series using the Flight service."""
        query = SeriesSearch(self.__source_name, selector.tags, selector.field)
        for result in self.__client.search(query):
            if isinstance(result, SeriesSelector):
                yield SeriesSelector.from_tags(
                    selector.source, result.tags, result.field
                )
            else:
                result.series = SeriesSelector.from_tags(
                    selector.source, result.series.tags, result.series.field
                )
                yield result

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Get metadata from the Flight service."""
        remote_selector = SeriesSelector.from_tags(
            self.__source_name, selector.tags, selector.field
        )
        metadata = self.__client.get_metadata(remote_selector)
        metadata.series = SeriesSelector(selector.source, selector.tags, selector.field)
        return metadata

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Get data from the Flight service."""
        remote_selector = SeriesSelector.from_tags(
            self.__source_name, selector.tags, selector.field
        )
        return self.__client.get_data(remote_selector, start_date, end_date)

    def get_plot_data(
        self,
        selector: SeriesSelector,
        start_date: datetime,
        end_date: datetime,
        interval_count: int,
    ) -> pa.Table:
        """Get plot data from the Flight service."""
        remote_selector = SeriesSelector.from_tags(
            self.__source_name, selector.tags, selector.field
        )
        return self.__client.get_plot_data(
            remote_selector, start_date, end_date, interval_count
        )

    def get_source_structure(
        self, selector: SeriesSelector
    ) -> Optional[SourceStructure]:
        """Return the source structure using the Flight service."""
        remote_selector = SeriesSelector.from_tags(
            self.__source_name, selector.tags, selector.field
        )
        return self.__client.get_source_structure(remote_selector)



"""Kukur source for PI Data Archives using PI Web API."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Generator, Optional

import pyarrow as pa

try:
    import urllib3
    from requests import Session

    HAS_REQUESTS = True
except ImportError:
    HAS_REQUESTS = False

from dateutil.parser import parse as parse_date

try:
    from requests_kerberos import HTTPKerberosAuth

    HAS_REQUESTS_KERBEROS = True
except ImportError:
    HAS_REQUESTS_KERBEROS = False

from kukur import (
    DataType,
    Dictionary,
    InterpolationType,
    Metadata,
    SeriesSearch,
    SeriesSelector,
)
from kukur.exceptions import (
    InvalidDataError,
    InvalidSourceException,
    MissingModuleException,
)
from kukur.metadata import fields


@dataclass
class _RequestProperties:

    verify_ssl: bool
    max_returned_items_per_call: int


class _DictionaryLookup:  # pylint: disable=too-few-public-methods
    def __init__(
        self,
        session,
        request_properties: _RequestProperties,
        data_archive: dict,
    ):
        self.__session = session
        self.__request_properties = request_properties
        self.__data_archive = data_archive
        self.__digital_set_links: Optional[Dict[str, str]] = None
        self.__dictionaries: Dict[str, Dictionary] = {}

    def get(self, name: str) -> Dictionary:
        """Return a dictionary with the given name.

        This caches dictionaries."""
        if name not in self.__dictionaries:
            self.__dictionaries[name] = self._get_dictionary(name)
        return self.__dictionaries[name]

    def _get_dictionary(self, name: str) -> Dictionary:
        if self.__digital_set_links is None:
            self.__digital_set_links = self._get_digital_set_links()
        response = self.__session.get(
            self.__digital_set_links[name],
            verify=self.__request_properties.verify_ssl,
            params=dict(maxCount=self.__request_properties.max_returned_items_per_call),
        )
        response.raise_for_status()

        return Dictionary(
            mapping={
                item["Value"]: item["Name"] for item in response.json().get("Items", [])
            }
        )

    def _get_digital_set_links(self) -> Dict[str, str]:
        response = self.__session.get(
            self.__data_archive["Links"]["EnumerationSets"],
            verify=self.__request_properties.verify_ssl,
            params=dict(maxCount=self.__request_properties.max_returned_items_per_call),
        )
        response.raise_for_status()

        return {
            item["Name"]: item["Links"]["Values"]
            for item in response.json().get("Items", [])
        }


class PIWebAPIDataArchiveSource:
    """Connect to PI Data Archives using the PI Web API."""

    def __init__(self, config: dict):
        self.__request_properties = _RequestProperties(
            verify_ssl=config.get("verify_ssl", True),
            max_returned_items_per_call=config.get(
                "max_returned_items_per_call", 150000
            ),
        )
        self.__data_archive_uri = config["data_archive_uri"]

        self.__basic_auth = None
        if "username" in config and "password" in config:
            self.__basic_auth = (config["username"], config["password"])

        if not self.__request_properties.verify_ssl:
            urllib3.disable_warnings()

    def search(self, selector: SeriesSearch) -> Generator[Metadata, None, None]:
        """Return all tags in the Data Archive."""
        session = self._get_session()

        response = session.get(
            self.__data_archive_uri,
            verify=self.__request_properties.verify_ssl,
            params=dict(selectedFields="Links.Points;Links.EnumerationSets"),
        )
        response.raise_for_status()

        data_archive = response.json()

        dictionary_lookup = _DictionaryLookup(
            session, self.__request_properties, data_archive
        )

        page = 0
        while True:
            response = session.get(
                data_archive["Links"]["Points"],
                verify=self.__request_properties.verify_ssl,
                params=dict(
                    maxCount=self.__request_properties.max_returned_items_per_call,
                    startIndex=page
                    * self.__request_properties.max_returned_items_per_call,
                ),
            )
            response.raise_for_status()
            points = response.json().get("Items", [])
            if len(points) == 0:
                break

            page = page + 1
            for point in points:
                metadata = _get_metadata(
                    SeriesSelector(selector.source, point["Name"]),
                    point,
                    dictionary_lookup,
                )
                if metadata is not None:
                    yield metadata

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Return metadata for one tag."""
        session = self._get_session()
        response = session.get(
            self.__data_archive_uri,
            verify=self.__request_properties.verify_ssl,
            params=dict(selectedFields="Links.Points;Links.EnumerationSets"),
        )
        response.raise_for_status()

        data_archive = response.json()
        response = session.get(
            data_archive["Links"]["Points"],
            verify=self.__request_properties.verify_ssl,
            params=dict(nameFilter=selector.name),
        )
        response.raise_for_status()

        dictionary_lookup = _DictionaryLookup(
            session, self.__request_properties, data_archive
        )

        items = response.json()["Items"]
        if len(items) == 0:
            raise InvalidDataError("Series not found")

        metadata = _get_metadata(selector, items[0], dictionary_lookup)
        if metadata is None:
            return Metadata(selector)
        return metadata

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Return data for the given time series in the given time period."""
        session = self._get_session()
        response = session.get(
            self._get_data_url(session, selector),
            verify=self.__request_properties.verify_ssl,
            params=dict(
                maxCount=str(self.__request_properties.max_returned_items_per_call),
                startTime=start_date.isoformat(),
                endTime=end_date.isoformat(),
            ),
        )

        data_points = response.json()["Items"]
        if len(data_points) == self.__request_properties.max_returned_items_per_call:
            raise InvalidDataError("Increase max_returned_items_per_call")

        timestamps = []
        values = []
        quality_flags = []
        for data_point in data_points:
            value = data_point["Value"]
            if isinstance(value, dict):
                if value.get("IsSystem", False):
                    continue
                values.append(value["Value"])
            else:
                values.append(value)
            timestamps.append(parse_date(data_point["Timestamp"]))
            if data_point["Good"]:
                quality_flags.append(1)
            else:
                quality_flags.append(0)

        return pa.Table.from_pydict(
            {"ts": timestamps, "value": values, "quality": quality_flags}
        )

    def _get_session(self):
        session = Session()
        if self.__basic_auth is None and HAS_REQUESTS_KERBEROS:
            session.auth = HTTPKerberosAuth(
                mutual_authentication="REQUIRED", sanitize_mutual_error_response=False
            )
        elif self.__basic_auth is not None:
            session.auth = self.__basic_auth
        return session

    def _get_data_url(self, session, selector: SeriesSelector) -> str:
        response = session.get(
            self.__data_archive_uri,
            verify=self.__request_properties.verify_ssl,
            params=dict(selectedFields="Links.Points"),
        )
        response.raise_for_status()

        data_archive = response.json()
        response = session.get(
            data_archive["Links"]["Points"],
            verify=self.__request_properties.verify_ssl,
            params=dict(
                nameFilter=selector.name,
                selectedFields="Items.Links.RecordedData",
            ),
        )
        response.raise_for_status()

        data_points = response.json()["Items"]
        if len(data_points) == 0:
            raise InvalidDataError("Series not found")

        return response.json()["Items"][0]["Links"]["RecordedData"]


def _get_metadata(
    selector: SeriesSelector, point: dict, dictionary_lookup: _DictionaryLookup
) -> Optional[Metadata]:
    metadata = Metadata(SeriesSelector(selector.source, point["Name"]))
    metadata.set_field(fields.Description, point["Descriptor"])
    metadata.set_field(fields.Unit, point["EngineeringUnits"])

    if point["Step"]:
        metadata.set_field(fields.InterpolationType, InterpolationType.STEPPED)
    else:
        metadata.set_field(fields.InterpolationType, InterpolationType.LINEAR)

    metadata.set_field(fields.LimitLowFunctional, point["Zero"])
    metadata.set_field(fields.LimitHighFunctional, point["Zero"] + point["Span"])

    if len(point["DigitalSetName"]) > 0:
        dictionary_name = point["DigitalSetName"]
        metadata.set_field(fields.DictionaryName, dictionary_name)
        metadata.set_field(fields.Dictionary, dictionary_lookup.get(dictionary_name))

    point_type = point["PointType"]
    point_types = {
        "Digital": DataType.DICTIONARY,
        "Float16": DataType.FLOAT32,
        "Float32": DataType.FLOAT32,
        "Float64": DataType.FLOAT64,
        "Int16": DataType.FLOAT32,
        "Int32": DataType.FLOAT64,
        "String": DataType.STRING,
    }
    if point_type not in point_types:
        return None
    metadata.set_field(fields.DataType, point_types[point_type])
    return metadata


def from_config(config: dict) -> PIWebAPIDataArchiveSource:
    """Create a new PIWebAPIDataArchiveSource."""
    if "data_archive_uri" not in config:
        raise InvalidSourceException(
            'piwebapi-da sources require a "data_archive_uri" entry'
        )
    if not HAS_REQUESTS:
        raise MissingModuleException("requests", "piwebapi-da")
    return PIWebAPIDataArchiveSource(config)



"""Persistence for api keys"""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
from typing import List, Tuple, Optional
from datetime import datetime
from kukur.repository.base import BaseRepository, Migration

from kukur.api_key import ApiKey


class ApiKeyMigration(Migration):  # pylint: disable=too-few-public-methods
    """Create database tables for api keys."""

    def migrate(self):
        with self._open_db() as db:
            cursor = db.cursor()
            cursor.executescript(
                """
                create table if not exists ApiKey (
                    id integer primary key autoincrement,
                    name text not null unique,
                    api_key blob not null,
                    salt blob not null,
                    creation_date datetime not null
                );
            """
            )
            db.commit()


class ApiKeyRepository(BaseRepository):
    """ApiKeyRepository stores and retrieves api keys."""

    def migrations(self):
        """Return the database migrations for the api key tables."""
        return ApiKeyMigration(self._get_connection_string())

    def store(self, name: str, api_key: bytes, salt: bytes, creation_date: datetime):
        """Create a new api key with the given name.

        It returns the newly created api_key."""
        with self._open_db() as db:
            cursor = db.cursor()
            cursor.execute(
                """
                insert into ApiKey (name, api_key, salt, creation_date)
                values (?, ?, ?, ?)
            """,
                [name, api_key, salt, creation_date],
            )
            db.commit()

    def list(self) -> List[ApiKey]:
        """List all api keys."""
        with self._open_db() as db:
            cursor = db.cursor()
            cursor.execute(
                """
            select name, creation_date
            from ApiKey
            """
            )
            results = cursor.fetchall()
            return [ApiKey(name, creation_date) for name, creation_date in results]

    def get(self, name) -> Tuple[Optional[bytes], Optional[bytes]]:
        """Get an api key by name."""
        with self._open_db() as db:
            cursor = db.cursor()
            cursor.execute(
                """
            select api_key, salt
            from ApiKey
            where name = ?
            """,
                [name],
            )
            result = cursor.fetchone()
            if result is None:
                return None, None
            api_key, salt = result
            return api_key, salt

    def has_api_key(self, name: str) -> bool:
        """Check if api key exists."""
        with self._open_db() as db:
            cursor = db.cursor()
            cursor.execute(
                """
            select count(*)
            from ApiKey
            where name = ?
            """,
                [name],
            )
            return cursor.fetchone()[0] == 1

    def revoke(self, name: str):
        """Delete an api key."""
        with self._open_db() as db:
            cursor = db.cursor()
            cursor.execute(
                """
                delete from ApiKey
                where name = ?
            """,
                [name],
            )
            db.commit()



"""Quality mapper for Kukur data sources."""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from enum import Enum
from typing import Any, Dict, List, Union


class Quality(Enum):
    """Enumeration of possible Kukur quality flags."""

    BAD = 0
    GOOD = 1


class QualityMapper:
    """QualityMapper maps quality values used in a source to the quality values of Kukur.
    GOOD = 1 and BAD = 0."""

    __good_mapping: List[Union[str, int]]

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "QualityMapper":
        """Create a new mapper from a dictionary that maps Kukur quality values to external quality values."""
        mapper = cls()
        for quality, quality_values in config.items():
            if quality == Quality.GOOD.name:
                for value_list in quality_values:
                    if isinstance(value_list, list):
                        if len(value_list) > 1:
                            mapper.add_mapping_range(
                                range(int(value_list[0]), int(value_list[1]) + 1)
                            )
                        mapper.add_mapping(value_list[0])
                        continue
                    mapper.add_mapping(value_list)
        return mapper

    def __init__(self):
        self.__good_mapping = []

    def add_mapping(self, quality_value: Union[str, int]):
        """Add a mapping"""
        self.__good_mapping.append(quality_value)

    def add_mapping_range(self, quality_values: range):
        """Add a mapping range"""
        self.__good_mapping.extend(quality_values)

    def from_source(self, source_quality_value: Union[str, int]) -> int:
        """Map a quality value of a source to the Kukur quality value."""
        if source_quality_value in self.__good_mapping:
            return Quality.GOOD.value
        return Quality.BAD.value

    def is_present(self) -> bool:
        """Check if there is a quality mapping present."""
        return len(self.__good_mapping) > 0



"""Connect to another Kukur instance using Arrow Flight."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
from .kukur import from_config

__all__ = ["from_config"]



"""Connections to azure data explorer data sources from Timeseer."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

import json

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Generator, List, Optional

import pyarrow as pa

try:
    from azure.identity import DefaultAzureCredential

    HAS_AZURE_IDENTITY = True
except ImportError:
    HAS_AZURE_IDENTITY = False

try:
    from azure.kusto.data import KustoClient, KustoConnectionStringBuilder

    HAS_KUSTO = True
except ImportError:
    HAS_KUSTO = False


from kukur import Metadata, SeriesSearch, SeriesSelector, SourceStructure
from kukur.exceptions import (
    KukurException,
    MissingModuleException,
)
from kukur.source.metadata import MetadataMapper, MetadataValueMapper


class InvalidClientConnection(KukurException):
    """Raised when an error occured when making the connection."""

    def __init__(self, message: str):
        super().__init__(self, f"Connection error: {message}")


@dataclass
class DataExplorerConfiguration:
    """Data Explorer source configuration."""

    connection_string: str
    database: str
    table: str
    timestamp_column: str
    tags: List[str]
    metadata_columns: List[str]
    ignored_columns: List[str]


def from_config(
    config: Dict[str, Any],
    metadata_mapper: MetadataMapper,
    metadata_value_mapper: MetadataValueMapper,
):
    """Create a new Azure Data Explorer data source"""
    connection_string = config["connection_string"]
    database = config["database"]
    table = config["table"]
    timestamp_column = config.get("timestamp_column", "ts")
    tags = config.get("tag_columns", [])
    metadata_columns = config.get("metadata_columns", [])
    ignored_columns = config.get("ignored_columns", [])
    return DataExplorerSource(
        DataExplorerConfiguration(
            connection_string,
            database,
            table,
            timestamp_column,
            tags,
            metadata_columns,
            ignored_columns,
        ),
        metadata_mapper,
        metadata_value_mapper,
    )


class DataExplorerSource:  # pylint: disable=too-many-instance-attributes
    """An Azure Data Explorer data source."""

    __database: str
    __table: str
    __timestamp_column: str
    __tags: List[str]
    __metadata_columns: List[str]
    __ignored_columns: List[str]
    __metadata_mapper: MetadataMapper
    __metadata_value_mapper: MetadataValueMapper

    if HAS_KUSTO:
        __client: KustoClient

    def __init__(
        self,
        config: DataExplorerConfiguration,
        metadata_mapper: MetadataMapper,
        metadata_value_mapper: MetadataValueMapper,
    ):
        if not HAS_AZURE_IDENTITY:
            raise MissingModuleException("azure-identity")

        if not HAS_KUSTO:
            raise MissingModuleException("azure-kusto-data", "data_explorer")

        self.__metadata_mapper = metadata_mapper
        self.__metadata_value_mapper = metadata_value_mapper
        self.__database = _escape(config.database)
        self.__table = _escape(config.table)
        self.__timestamp_column = _escape(config.timestamp_column)
        self.__tags = [_escape(tag) for tag in config.tags]
        self.__metadata_columns = [
            _escape(metadata_column) for metadata_column in config.metadata_columns
        ]
        self.__ignored_columns = [
            _escape(ignored_column) for ignored_column in config.ignored_columns
        ]

        azure_credential = DefaultAzureCredential()

        def _get_token():
            return azure_credential.get_token(config.connection_string + "//.default")[
                0
            ]

        kcsb = KustoConnectionStringBuilder.with_token_provider(
            config.connection_string, _get_token
        )
        self.__client = KustoClient(kcsb)

    def search(self, selector: SeriesSearch) -> Generator[Metadata, None, None]:
        """Search for series matching the given selector."""
        if len(self.__tags) == 0:
            raise KukurException("Define tags to support listing time series")

        all_columns = {_escape(column) for column in self._get_table_columns()}
        fields = (
            all_columns
            - set([self.__timestamp_column])
            - set(self.__tags)
            - set(self.__metadata_columns)
            - set(self.__ignored_columns)
        )

        if len(self.__metadata_columns) == 0:
            query = f"['{self.__table}'] | distinct {', '.join(self.__tags)}"
            result = self.__client.execute(self.__database, query)
            if result is None or len(result.primary_results) == 0:
                return

            for row in result.primary_results[0]:
                tags = {}
                for tag in self.__tags:
                    tags[tag] = row[tag]
                for field in fields:
                    yield Metadata(SeriesSelector(selector.source, tags, field))
        else:
            summaries = [
                f"['{name}']=arg_max(['{self.__timestamp_column}'], ['{name}'])"
                for name in self.__metadata_columns
            ]
            renames = [f"['{name}']=['{name}1']" for name in self.__metadata_columns]
            query = f"""['{self.__table}']
                | summarize {', '.join(summaries)} by {', '.join(_add_square_brackets(self.__tags))}
                | project-away {', '.join(_add_square_brackets(self.__metadata_columns))}
                | project-rename {', '.join(renames)}
            """
            result = self.__client.execute(self.__database, query)
            if result is None or len(result.primary_results) == 0:
                return

            for row in result.primary_results[0]:
                tags = {}
                for tag in self.__tags:
                    tags[tag] = row[tag]
                for field in fields:
                    series_selector = SeriesSelector(selector.source, tags, field)
                    metadata = Metadata(series_selector)
                    for column_name in self.__metadata_columns:
                        metadata.coerce_field(
                            self.__metadata_mapper.from_source(column_name),
                            self.__metadata_value_mapper.from_source(
                                self.__metadata_mapper.from_source(column_name),
                                row[column_name],
                            ),
                        )
                    yield metadata

    # pylint: disable=no-self-use
    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Data explorer currently always returns empty metadata."""
        return Metadata(selector)

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Return data for the given time series in the given time period."""

        query = f"""['{self.__table}']
            | where ['{self.__timestamp_column}'] >= todatetime('{start_date}')
            | where ['{self.__timestamp_column}'] <= todatetime('{end_date}')
        """

        for (tag_key, tag_value) in selector.tags.items():
            query += f" | where ['{_escape(tag_key)}']=='{_escape(tag_value)}'"

        query = f"{query} | sort by ['{self.__timestamp_column}'] asc"

        result = self.__client.execute(self.__database, query)
        timestamps = []
        values = []

        if result is not None and len(result.primary_results) > 0:
            for row in result.primary_results[0]:
                timestamps.append(row[self.__timestamp_column])
                values.append(row[selector.field])

        return pa.Table.from_pydict({"ts": timestamps, "value": values})

    def get_source_structure(self, _: SeriesSelector) -> Optional[SourceStructure]:
        """Return the available tag keys, tag values and tag fields."""
        all_columns = {_escape(column) for column in self._get_table_columns()}
        fields = (
            all_columns
            - set([self.__timestamp_column])
            - set(self.__tags)
            - set(self.__metadata_columns)
            - set(self.__ignored_columns)
        )
        tag_keys = set(self.__tags)
        tag_values = []

        for tag_key in tag_keys:
            query = f"""['{self.__table}']
                    | distinct ['{tag_key}']
            """
            result = self.__client.execute(self.__database, query)
            if result is not None and len(result.primary_results) > 0:
                tag_values.extend(
                    [
                        {"key": tag_key, "value": row[0]}
                        for row in result.primary_results[0]
                    ]
                )

        return SourceStructure(list(fields), list(tag_keys), tag_values)

    def _get_table_columns(self) -> List[str]:
        query = f".show table ['{self.__table}'] schema as json"
        result = self.__client.execute(self.__database, query)
        if result is None or len(result.primary_results) == 0:
            raise KukurException("Failed to get table schema")
        for row in result.primary_results[0]:
            schema = json.loads(row["Schema"])
            return [column["Name"] for column in schema.get("OrderedColumns", [])]
        raise KukurException("Table schema is empty")


def _add_square_brackets(columns: List[str]) -> List[str]:
    return [f"['{column}']" for column in columns]


def _escape(context: Optional[str]) -> str:
    if context is None:
        context = "value"
    if "'" in context:
        context = context.replace("'", "")
    if '"' in context:
        context = context.replace('"', "")
    if "|" in context:
        context = context.replace("|", "")
    if ";" in context:
        context = context.replace(";", "")
    return context



"""Connections to ADODB data sources from Timeseer.

This requires an installation of pywin32 (LGPL).
"""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
from .adodb import from_config

__all__ = ["from_config"]



"""Test flexible Metadata"""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from kukur import (
    DataType,
    Dictionary,
    InterpolationType,
    SeriesSelector,
)
from kukur.metadata import Metadata, MetadataField, fields

SERIES = SeriesSelector("test", "test-tag-1")


def test_series_json() -> None:
    metadata = Metadata(SERIES)
    data = metadata.to_data()
    assert data["series"]["source"] == "test"
    assert data["series"]["tags"]["series name"] == "test-tag-1"

    new_metadata = Metadata.from_data(data)
    assert new_metadata.series == SeriesSelector(
        SERIES.source, SERIES.tags, SERIES.field
    )


def test_repr() -> None:
    metadata = Metadata(SERIES)
    metadata.set_field(fields.Description, "a test tag")
    metadata.set_field(fields.InterpolationType, InterpolationType.LINEAR)
    metadata.set_field_by_name("custom", "value")
    metadata.set_field_by_name("custom2", "value2")

    assert (
        "SeriesSelector(source='test', tags={'series name': 'test-tag-1'}, field='value')"
        in repr(metadata)
    )
    assert "'description': 'a test tag'" in repr(metadata)
    assert "'interpolation type': <InterpolationType.LINEAR: 'LINEAR'>" in repr(
        metadata
    )
    assert "'custom': 'value'" in repr(metadata)
    assert "'custom2': 'value2'" in repr(metadata)


def test_typed_field() -> None:
    CustomField = MetadataField[str]("custom", default="", serialized_name="custom")
    Metadata.register_field(CustomField, after_field=fields.Description)

    metadata = Metadata(SERIES)
    metadata.set_field(CustomField, "test")
    assert list(metadata.iter_names())[1] == ("custom", "test")


def test_description() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.Description) == ""

    metadata.set_field(fields.Description, "a test tag")
    assert metadata.get_field(fields.Description) == "a test tag"


def test_description_json() -> None:
    metadata = Metadata(SERIES, {fields.Description: "a test tag"})
    data = metadata.to_data()
    assert data["description"] == "a test tag"

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.Description) == "a test tag"


def test_unit() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.Unit) == ""

    metadata.set_field(fields.Unit, "kg")
    assert metadata.get_field(fields.Unit) == "kg"


def test_unit_json() -> None:
    metadata = Metadata(SERIES, {fields.Unit: "kg"})
    data = metadata.to_data()
    assert data["unit"] == "kg"

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.Unit) == "kg"


def test_physical_limit_low() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.LimitLowPhysical) is None

    metadata.set_field(fields.LimitLowPhysical, 0)
    assert metadata.get_field(fields.LimitLowPhysical) == 0


def test_physical_limit_low_json() -> None:
    metadata = Metadata(SERIES, {fields.LimitLowPhysical: 0})
    data = metadata.to_data()
    assert data["limitLowPhysical"] == 0

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.LimitLowPhysical) == 0


def test_physical_limit_low_coerce() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("physical lower limit", "0")
    assert metadata.get_field(fields.LimitLowPhysical) == 0


def test_physical_limit_high() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.LimitHighPhysical) is None

    metadata.set_field(fields.LimitHighPhysical, 0)
    assert metadata.get_field(fields.LimitHighPhysical) == 0


def test_physical_limit_high_json() -> None:
    metadata = Metadata(SERIES, {fields.LimitHighPhysical: 0})
    data = metadata.to_data()
    assert data["limitHighPhysical"] == 0

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.LimitHighPhysical) == 0


def test_physical_limit_high_coerce() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("physical upper limit", "0")
    assert metadata.get_field(fields.LimitHighPhysical) == 0


def test_functional_limit_low() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.LimitLowFunctional) is None

    metadata.set_field(fields.LimitLowFunctional, 0)
    assert metadata.get_field(fields.LimitLowFunctional) == 0


def test_functional_limit_low_json() -> None:
    metadata = Metadata(SERIES, {fields.LimitLowFunctional: 0})
    data = metadata.to_data()
    assert data["limitLowFunctional"] == 0

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.LimitLowFunctional) == 0


def test_functional_limit_low_coerce() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("functional lower limit", "0")
    assert metadata.get_field(fields.LimitLowFunctional) == 0


def test_functional_limit_high() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.LimitHighFunctional) is None

    metadata.set_field(fields.LimitHighFunctional, 0)
    assert metadata.get_field(fields.LimitHighFunctional) == 0


def test_functional_limit_high_json() -> None:
    metadata = Metadata(SERIES, {fields.LimitHighFunctional: 0})
    data = metadata.to_data()
    assert data["limitHighFunctional"] == 0

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.LimitHighFunctional) == 0


def test_functional_limit_high_coerce() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("functional upper limit", "0")
    assert metadata.get_field(fields.LimitHighFunctional) == 0


def test_accuracy() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.Accuracy) is None

    metadata.set_field(fields.Accuracy, 0.2)
    assert metadata.get_field(fields.Accuracy) == 0.2


def test_accuracy_json() -> None:
    metadata = Metadata(SERIES, {fields.Accuracy: 0.2})
    data = metadata.to_data()
    assert data["accuracy"] == 0.2

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.Accuracy) == 0.2


def test_accuracy_coerce() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("accuracy", "0.2")
    assert metadata.get_field(fields.Accuracy) == 0.2


def test_accuracy_percentage() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.Accuracy) is None

    metadata.set_field(fields.AccuracyPercentage, 2)
    metadata.set_field(fields.LimitLowPhysical, 0)
    metadata.set_field(fields.LimitHighPhysical, 10)
    assert metadata.get_field(fields.Accuracy) == 0.2
    assert metadata.get_field(fields.AccuracyPercentage) == 2


def test_accuracy_percentage_outside_range() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.Accuracy) is None

    metadata.set_field(fields.AccuracyPercentage, 200)
    metadata.set_field(fields.LimitLowPhysical, 0)
    metadata.set_field(fields.LimitHighPhysical, 10)
    assert metadata.get_field(fields.Accuracy) is None
    assert metadata.get_field(fields.AccuracyPercentage) == 200


def test_accuracy_percentage_with_accuracy() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.Accuracy) is None

    metadata.set_field(fields.Accuracy, 1)
    metadata.set_field(fields.AccuracyPercentage, 2)
    metadata.set_field(fields.LimitLowPhysical, 0)
    metadata.set_field(fields.LimitHighPhysical, 10)
    assert metadata.get_field(fields.Accuracy) == 1
    assert metadata.get_field(fields.AccuracyPercentage) == 2


def test_accuracy_percentage_no_physical_limits() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.Accuracy) is None

    metadata.set_field(fields.AccuracyPercentage, 2)
    metadata.set_field(fields.LimitLowFunctional, 0)
    metadata.set_field(fields.LimitHighFunctional, 10)
    assert metadata.get_field(fields.Accuracy) == 0.2
    assert metadata.get_field(fields.AccuracyPercentage) == 2


def test_accuracy_percentage_no_limits() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.Accuracy) is None

    metadata.set_field(fields.AccuracyPercentage, 2)
    assert metadata.get_field(fields.Accuracy) is None
    assert metadata.get_field(fields.AccuracyPercentage) == 2


def test_percentage_accuracy_json() -> None:
    metadata = Metadata(
        SERIES,
        {
            fields.AccuracyPercentage: 2,
            fields.LimitLowPhysical: 0,
            fields.LimitHighPhysical: 10,
        },
    )
    data = metadata.to_data()
    assert data["accuracyPercentage"] == 2

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.Accuracy) == 0.2
    assert new_metadata.get_field(fields.AccuracyPercentage) == 2


def test_percentage_accuracy_json_outside_range() -> None:
    metadata = Metadata(
        SERIES,
        {
            fields.AccuracyPercentage: 200,
            fields.LimitLowPhysical: 0,
            fields.LimitHighPhysical: 10,
        },
    )
    data = metadata.to_data()
    assert data["accuracyPercentage"] == 200

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.Accuracy) is None
    assert new_metadata.get_field(fields.AccuracyPercentage) == 200


def test_percentage_accuracy_json_with_accuracy() -> None:
    metadata = Metadata(
        SERIES,
        {
            fields.Accuracy: 1,
            fields.AccuracyPercentage: 2,
            fields.LimitLowPhysical: 0,
            fields.LimitHighPhysical: 10,
        },
    )
    data = metadata.to_data()
    assert data["accuracyPercentage"] == 2
    assert data["accuracy"] == 1

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.Accuracy) == 1
    assert metadata.get_field(fields.AccuracyPercentage) == 2


def test_percentage_accuracy_coerce() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("accuracy percentage", "2")
    metadata.coerce_field("physical lower limit", "0")
    metadata.coerce_field("physical upper limit", "10")
    assert metadata.get_field(fields.Accuracy) == 0.2


def test_percentage_accuracy_coerce_outside_range() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("accuracy percentage", "200")
    metadata.coerce_field("physical lower limit", "0")
    metadata.coerce_field("physical upper limit", "10")
    assert metadata.get_field(fields.Accuracy) is None
    assert metadata.get_field(fields.AccuracyPercentage) == 200


def test_percentage_accuracy_coerce_with_accuracy() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("accuracy", "1")
    metadata.coerce_field("accuracy percentage", "2")
    metadata.coerce_field("physical lower limit", "0")
    metadata.coerce_field("physical upper limit", "10")
    assert metadata.get_field(fields.Accuracy) == 1
    assert metadata.get_field(fields.AccuracyPercentage) == 2


def test_interpolation_type() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.InterpolationType) is None

    metadata.set_field(fields.InterpolationType, InterpolationType.LINEAR)
    assert metadata.get_field(fields.InterpolationType) == InterpolationType.LINEAR


def test_interpolation_type_json() -> None:
    metadata = Metadata(SERIES, {fields.InterpolationType: InterpolationType.LINEAR})
    data = metadata.to_data()
    assert data["interpolationType"] == "LINEAR"

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.InterpolationType) == InterpolationType.LINEAR


def test_interpolation_type_coerce() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("interpolation type", "STEPPED")
    assert metadata.get_field(fields.InterpolationType) == InterpolationType.STEPPED


def test_data_type() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.DataType) is None

    metadata.set_field(fields.DataType, DataType.STRING)
    assert metadata.get_field(fields.DataType) == DataType.STRING


def test_data_type_json() -> None:
    metadata = Metadata(SERIES, {fields.DataType: DataType.STRING})
    data = metadata.to_data()
    assert data["dataType"] == "STRING"

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.DataType) == DataType.STRING


def test_data_type_coerce() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("data type", "DICTIONARY")
    assert metadata.get_field(fields.DataType) == DataType.DICTIONARY


def test_dictionary_name() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.DictionaryName) is None

    metadata.set_field(fields.DictionaryName, "onoff")
    assert metadata.get_field(fields.DictionaryName) == "onoff"


def test_dictionary_name_json() -> None:
    metadata = Metadata(SERIES, {fields.DictionaryName: "onoff"})
    data = metadata.to_data()
    assert data["dictionaryName"] == "onoff"

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.DictionaryName) == "onoff"


def test_dictionary() -> None:
    dictionary = Dictionary({0: "OFF", 1: "ON"})

    metadata = Metadata(SERIES)
    assert metadata.get_field(fields.Dictionary) is None

    metadata.set_field(fields.Dictionary, dictionary)
    assert metadata.get_field(fields.Dictionary) == dictionary


def test_dictionary_json() -> None:
    dictionary = Dictionary({0: "OFF", 1: "ON"})

    metadata = Metadata(SERIES, {fields.Dictionary: dictionary})
    data = metadata.to_data()
    assert data["dictionary"] == [(0, "OFF"), (1, "ON")]

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field(fields.Dictionary) == dictionary


def test_dictionary_coerce() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("dictionary", [(0, "OFF"), (1, "ON")])
    assert metadata.get_field(fields.Dictionary) == Dictionary({0: "OFF", 1: "ON"})


def test_unknown() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("process type", "BATCH")

    assert metadata.get_field_by_name("process type") == "BATCH"


def test_unknown_iter() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("process type", "BATCH")

    assert dict(metadata.iter_serialized())["process type"] == "BATCH"


def test_unknown_json() -> None:
    metadata = Metadata(SERIES)
    metadata.coerce_field("process type", "BATCH")

    data = metadata.to_data()
    assert data["process type"] == "BATCH"

    new_metadata = Metadata.from_data(data, SERIES)
    assert new_metadata.get_field_by_name("process type") == "BATCH"


def test_non_existent_unknown() -> None:
    metadata = Metadata(SERIES)
    assert metadata.get_field_by_name("process type") is None



"""Expose Kukur using Arrow Flight."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
import json

from typing import Any, Callable, Dict, Generator, List

import pyarrow.flight as fl

from dateutil.parser import parse as parse_date

from kukur import PlotSource, SeriesSelector, Source, TagSource
from kukur.app import Kukur
from kukur.exceptions import InvalidSourceException

__all__ = ["JSONFlightServer"]


class JSONFlightServer(fl.FlightServerBase):
    """JSONFlightServer handles JSON Apache Arrow Flight tickets.

    Extra keyword arguments are passed to FlightServerBase.

    It supports registering custom actions and request handlers. Register a GET
    handler to return Arrow data. To return JSON, register an action handler."""

    __get_handlers: Dict[str, Callable]
    __action_handlers: Dict[str, Callable]

    def __init__(self, config: Dict[str, Any], **kwargs):
        host = "0.0.0.0"
        port = 8081
        if "flight" in config:
            host = config["flight"].get("host", host)
            port = config["flight"].get("port", port)
        super().__init__(location=(host, port), **kwargs)
        self.__get_handlers = {}
        self.__action_handlers = {}

    def register_get_handler(self, handler_name: str, func: Callable):
        """Register a handler for GET requests that returns Arrow data.

        The 'query' field in a request maps to the handler_name.

        The handler func will receive the Flight Context and the JSON-parsed ticket as arguments.
        """
        self.__get_handlers[handler_name] = func

    def register_action_handler(self, action_type: str, func: Callable):
        """Register a handler for action requests.

        The handler func will receive the Flight Context and the Flight Action.

        Note that there is no requirement to return JSON here."""
        self.__action_handlers[action_type] = func

    def do_get(self, context, ticket: fl.Ticket):
        """Respond with Arrow columnar data to the given ticket."""
        request = json.loads(ticket.ticket)

        return self.__get_handlers[request["query"]](context, request)

    def do_action(self, context, action: fl.Action):
        """Respond to a generic action request."""
        return self.__action_handlers[action.type](context, action)


class KukurFlightServer:
    """KukurFlightServer exposes the data sources provided by a SourceFactory over Arrow Flight."""

    def __init__(self, source: Source):
        self.__source = source

    def search(self, _, action: fl.Action) -> Generator[bytes, None, None]:
        """Search a data source for time series.

        This returns either a SeriesSelector or Metadata as JSON, depending on
        what is supported by the source."""
        request = json.loads(action.body.to_pybytes())
        selector = SeriesSelector.from_data(request)
        for result in self.__source.search(selector):
            yield json.dumps(result.to_data()).encode()

    def get_metadata(self, _, action: fl.Action) -> List[bytes]:
        """Return metadata for the given time series as JSON."""
        request = json.loads(action.body.to_pybytes())
        selector = SeriesSelector.from_data(request)
        metadata = self.__source.get_metadata(selector).to_data()
        return [json.dumps(metadata).encode()]

    def get_data(self, _, request) -> Any:
        """Return time series data as Arrow data."""
        selector = SeriesSelector.from_data(request["selector"])
        start_date = parse_date(request["start_date"])
        end_date = parse_date(request["end_date"])
        data = self.__source.get_data(selector, start_date, end_date)
        return fl.RecordBatchStream(data)

    def get_plot_data(self, _, request) -> Any:
        """Return plot data as Arrow."""
        selector = SeriesSelector.from_data(request["selector"])
        start_date = parse_date(request["start_date"])
        end_date = parse_date(request["end_date"])
        interval_count: int = request["interval_count"]
        if not isinstance(self.__source, PlotSource):
            raise InvalidSourceException("get_plot_data not supported by source")
        data = self.__source.get_plot_data(
            selector, start_date, end_date, interval_count
        )
        return fl.RecordBatchStream(data)

    def get_source_structure(self, _, action: fl.Action) -> List[bytes]:
        """Return the structure of a source for the given time series as JSON."""
        request = json.loads(action.body.to_pybytes())
        selector = SeriesSelector.from_data(request)
        if not isinstance(self.__source, TagSource):
            raise InvalidSourceException("get_source_structure not supported by source")
        source_structure = self.__source.get_source_structure(selector)
        if source_structure is None:
            return [json.dumps(source_structure).encode()]
        return [json.dumps(source_structure.to_data()).encode()]


class KukurServerAuthHandler(fl.ServerAuthHandler):
    """KukurServerAuthHandler handles the authentication"""

    _app: Kukur

    def __init__(self, app: Kukur):
        super().__init__()
        self._app = app

    def authenticate(self, outgoing, incoming):  # pylint: disable=no-self-use
        """Check the authentication."""
        buf = incoming.read()
        outgoing.write(buf)

    def is_valid(self, token: bytes):
        """Check if the supplied token is valid."""
        if token == "" or token is None:
            raise fl.FlightUnauthenticatedError("invalid token")

        auth = fl.BasicAuth.deserialize(token)
        if auth.username is None:
            raise fl.FlightUnauthenticatedError("invalid username")
        if auth.password is None:
            raise fl.FlightUnauthenticatedError("invalid password")

        if not self._app.get_api_keys().has_api_key(auth.username.decode("UTF-8")):
            raise fl.FlightUnauthenticatedError("invalid token")
        if not self._app.get_api_keys().is_valid(
            auth.username.decode("UTF-8"), auth.password.decode("UTF-8")
        ):
            raise fl.FlightUnauthenticatedError("invalid token")
        return token


class KukurServerNoAuthHandler(fl.ServerAuthHandler):
    """KukurServerNoAuthHandler handles the authentication when it is disabled"""

    def authenticate(self, outgoing, incoming):
        """Do nothing."""

    def is_valid(self, token: bytes):  # pylint: disable=no-self-use, unused-argument
        """Return empty string."""
        return ""



"""Simulate a data source by generating data for Timeseer."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from .simulator import from_config

__all__ = ["from_config"]



"""Dummy source for integration tests."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from .integration_test_source import from_config

__all__ = ["from_config"]



"""Service layer for api keys"""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
import secrets
import os

from hashlib import scrypt
from datetime import datetime
from typing import List

from kukur.api_key import ApiKey
from kukur.repository import RepositoryRegistry


class ApiKeys:
    """Api keys for authentication"""

    __repository: RepositoryRegistry

    def __init__(self, repository_registry: RepositoryRegistry):
        self.__repository = repository_registry

    def create(self, name: str) -> str:
        """Create an api key"""
        api_key = _create_random_api_key()
        creation_date = datetime.now()
        salt = os.urandom(16)
        self.__repository.api_key().store(
            name, _hash_api_key(api_key, salt), salt, creation_date
        )
        return api_key

    def list(self) -> List[ApiKey]:
        """List all the api keys"""
        return self.__repository.api_key().list()

    def has_api_key(self, name: str):
        """Get the api key by name"""
        return self.__repository.api_key().has_api_key(name)

    def is_valid(self, name: str, api_key: str) -> bool:
        """Check if the supplied api key is a valid one"""
        stored_api_key, salt = self.__repository.api_key().get(name)
        if salt is None or stored_api_key is None:
            return False
        hashed_api_key = _hash_api_key(api_key, salt)
        return stored_api_key == hashed_api_key

    def revoke(self, name: str) -> ApiKey:
        """Revoke an api key"""
        return self.__repository.api_key().revoke(name)


def _create_random_api_key() -> str:
    return secrets.token_urlsafe(40)


def _hash_api_key(api_key: str, salt: bytes) -> bytes:
    hashed = scrypt(bytes(api_key, "UTF-8"), salt=salt, n=16384, r=8, p=1)
    return hashed



"""Connections to InfluxDB data sources from Timeseer."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0

import json

from datetime import datetime
from typing import Any, Dict, Generator, Optional, Tuple

import dateutil.parser
import pyarrow as pa

try:
    from influxdb import InfluxDBClient
    from influxdb.exceptions import InfluxDBClientError

    HAS_INFLUX = True
except ImportError:
    HAS_INFLUX = False

from kukur import Metadata, SeriesSearch, SeriesSelector, SourceStructure
from kukur.exceptions import InvalidDataError, KukurException, MissingModuleException


class InvalidClientConnection(KukurException):
    """Raised when an error occured when making the connection."""

    def __init__(self, message: str):
        KukurException.__init__(self, f"Connection error: {message}")


def from_config(config: Dict[str, Any]):
    """Create a new Influx data source"""
    if not HAS_INFLUX:
        raise MissingModuleException("influxdb", "influxdb")
    host = config.get("host", "localhost")
    port = config.get("port", 8086)
    ssl = config.get("ssl", False)
    database = config["database"]
    username = config.get("username", "")
    password = config.get("password", "")
    return InfluxSource(host, port, ssl, database, username=username, password=password)


class InfluxSource:
    """An InfluxDB data source."""

    if HAS_INFLUX:
        __client: InfluxDBClient

    def __init__(
        self,
        host: str,
        port: int,
        ssl: bool,
        database: str,
        *,
        username: str,
        password: str,
    ):
        if not HAS_INFLUX:
            raise MissingModuleException("influxdb", "influxdb")
        try:
            if username != "" and password != "":
                self.__client = InfluxDBClient(
                    host=host, port=port, ssl=ssl, username=username, password=password
                )
            else:
                self.__client = InfluxDBClient(host=host, port=port, ssl=ssl)

            self.__client.switch_database(database)
        except InfluxDBClientError as err:
            raise InvalidClientConnection(err) from err

    def search(self, selector: SeriesSearch) -> Generator[Metadata, None, None]:
        """Search for series matching the given selector."""
        measurement = None
        if "series name" in selector.tags:
            measurement = selector.tags["series name"]
            del selector.tags["series name"]
        many_series = self.__client.get_list_series(
            measurement=measurement, tags=selector.tags
        )
        fields = self.__client.query("SHOW FIELD KEYS")
        for series in many_series:
            measurement, tags = _parse_influx_series(series)
            for field in fields.get_points(measurement=measurement):
                if selector.field in ["value", field["fieldKey"]]:
                    yield Metadata(
                        SeriesSelector.from_tags(
                            selector.source,
                            tags,
                            field["fieldKey"],
                        )
                    )

    # pylint: disable=no-self-use
    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Influx currently always returns empty metadata."""
        return Metadata(selector)

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Return data for the given time series in the given time period."""
        if "series name" not in selector.tags:
            raise InvalidDataError("No series name")

        query = f"""SELECT time, "{_escape(selector.field)}"
                    FROM "{_escape(selector.tags["series name"])}"
                    WHERE time >= $start_date and time <= $end_date"""

        bind_params = {
            "start_date": start_date.strftime("%Y-%m-%dT%H:%M:%SZ"),
            "end_date": end_date.strftime("%Y-%m-%dT%H:%M:%SZ"),
        }
        for i, (tag_key, tag_value) in enumerate(selector.tags.items()):
            if tag_key == "series name":
                continue
            bind_params[str(i)] = tag_value
            query = query + f' and "{_escape(tag_key)}" = ${str(i)}'

        timestamps = []
        values = []

        for item in self.__client.query(
            query=query, bind_params=bind_params
        ).get_points():
            timestamps.append(dateutil.parser.parse(item["time"]))
            values.append(item[selector.field])

        return pa.Table.from_pydict({"ts": timestamps, "value": values})

    def get_source_structure(self, _: SeriesSelector) -> Optional[SourceStructure]:
        """Return the available tag keys, tag value and tag fields."""
        query_tag_keys = "SHOW TAG KEYS"
        tag_keys = []
        for results in self.__client.query(query=query_tag_keys).get_points():
            tag_keys.extend(list(results.values()))
        tag_keys = list(set(tag_keys))

        query_fields = "SHOW FIELD KEYS"
        fields = []
        for results in self.__client.query(query=query_fields).get_points():
            for key, value in results.items():
                if key == "fieldKey":
                    fields.append(value)

        tag_key_placeholder = json.dumps(tag_keys)
        query_tag_values = f'SHOW TAG VALUES WITH KEY IN {tag_key_placeholder.replace("[", "(").replace("]", ")")}'
        tag_values = []
        for result in self.__client.query(query=query_tag_values).get_points():
            if result not in tag_values:
                tag_values.append(result)
        return SourceStructure(fields, tag_keys, tag_values)


def _parse_influx_series(series: str) -> Tuple[str, Dict[str, str]]:
    series_name = series.replace("\\", "")
    measurement = series_name.split(",")[0]
    tags = {}
    for tag in series_name.split(","):
        split_tag = tag.split("=")
        if len(split_tag) == 1:
            tags["series name"] = split_tag[0]
        else:
            tags[split_tag[0]] = split_tag[1]
    return measurement, tags


def _escape(context: Optional[str]) -> str:
    if context is None:
        context = "value"
    if '"' in context:
        context = context.replace('"', "")
    if ";" in context:
        context = context.replace(";", "")
    return context



"""Connections to azure data explorer data sources from Timeseer."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from .azure_data_explorer import from_config

__all__ = ["from_config"]



"""Simulate a data source by generating data for Timeseer."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

import itertools
import operator
import sys

from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from hashlib import sha1
from pathlib import Path
from random import Random
from typing import Any, Dict, Generator, List, Optional, Union

import numpy

import pyarrow as pa

try:
    import yaml

    HAS_YAML = True
except ImportError:
    HAS_YAML = False

from kukur import (
    Metadata,
    SeriesSearch,
    SeriesSelector,
    SignalGenerator,
    SourceStructure,
)
from kukur.exceptions import MissingModuleException


operator_functions = {
    "+": operator.add,
    "-": operator.sub,
}


@dataclass
class SimulatorConfiguration:
    """Simulator source configuration."""

    path: Optional[str]


def from_config(
    config: Dict[str, Any],
):
    """Create a new Simulator data source from the given configuration dictionary."""
    return SimulatorSource(SimulatorConfiguration(config.get("path")))


@dataclass
class SignalGeneratorConfig:
    """Base signal generator configuration."""

    series_name: str
    signal_type: str
    initial_seed: int
    number_of_seeds: int
    interval_seconds_min: Union[List[float], float]
    interval_seconds_max: Union[List[float], float]
    metadata: Dict[str, str]


@dataclass
class SignalConfig:
    """Base signal configuration."""

    series_name: str
    signal_type: str
    seed: int
    interval_seconds_min: int
    interval_seconds_max: int

    def to_bytes(self) -> bytes:
        """Convert to a unique bytes representation."""
        return bytes(str(self), "UTF-8")


@dataclass
class StepSignalGeneratorConfigValue:
    """One possible step configuration."""

    min: float
    max: float
    number_of_steps: List[int]


@dataclass
class StepSignalGeneratorConfig(SignalGeneratorConfig):
    """Configuration for the step signal generator."""

    values: List[StepSignalGeneratorConfigValue]


@dataclass
class StepSignalConfig(SignalConfig):
    """Configuration for the step signal."""

    min_value: float
    max_value: float
    number_of_steps: int


@dataclass
class WhiteNoiseSignalGeneratorConfig(SignalGeneratorConfig):
    """Configuration for the white noise signal generator."""

    mean: Union[List[float], float]
    standard_deviation: Union[List[float], float]


@dataclass
class WhiteNoiseSignalConfig(SignalConfig):
    """Configuration for the white noise signal."""

    mean: float
    standard_deviation: float


@dataclass
class SineSignalGeneratorConfig(SignalGeneratorConfig):
    """Configuration for the sine signal generator."""

    period_seconds: Union[List[float], float]
    phase_seconds: Union[List[float], float]
    amplitude: Union[List[float], float]
    shift: Union[List[float], float]


@dataclass
class SineSignalConfig(SignalConfig):
    """Configuration for one sine signal."""

    period_seconds: float
    phase_seconds: float
    amplitude: float
    shift: float


class StepSignalGenerator:
    """Step signal generator."""

    __config: Optional[StepSignalGeneratorConfig] = None

    def __init__(self, config: Optional[Dict] = None):
        if config is not None:
            self.__config = StepSignalGeneratorConfig(
                config["seriesName"],
                config["type"],
                config.get("initialSeed", 0),
                config.get("numberOfSeeds", 1),
                config["samplingInterval"]["intervalSecondsMin"],
                config["samplingInterval"]["intervalSecondsMax"],
                config.get("metadata", {}),
                [
                    StepSignalGeneratorConfigValue(
                        value["min"], value["max"], _ensure_list(value["numberOfSteps"])
                    )
                    for value in config["values"]
                ],
            )

    def generate(  # pylint: disable=no-self-use, too-many-locals
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Generates data in steps based on a selector start and end date."""
        current_time = _get_start_of_day(start_date)
        configuration = _get_step_configuration(selector)
        ts = []
        value = []
        rng = Random(_get_hex_digest(current_time, configuration.to_bytes()))

        step_size = (
            configuration.max_value - configuration.min_value
        ) / configuration.number_of_steps
        current_value = (
            configuration.min_value
            + rng.randint(0, configuration.number_of_steps) * step_size
        )
        while current_time <= end_date:
            generated_step = (
                rng.randint(0, int(configuration.number_of_steps / 2)) * step_size
            )
            random_operator = rng.choice(["+", "-"])
            new_value = operator_functions[random_operator](
                current_value, generated_step
            )
            if (
                new_value > configuration.max_value
                or new_value < configuration.min_value
            ):
                random_operator = "+" if random_operator == "-" else "-"
                new_value = operator_functions[random_operator](
                    current_value, generated_step
                )

            value.append(new_value)
            ts.append(current_time)
            current_value = new_value

            time_increment = rng.randint(
                configuration.interval_seconds_min, configuration.interval_seconds_max
            )
            new_time = current_time + timedelta(seconds=time_increment)

            if new_time.date() != current_time.date():
                new_time = _get_start_of_day(new_time)
                rng.seed(_get_hex_digest(new_time, configuration.to_bytes()))
                current_value = (
                    configuration.min_value
                    + rng.randint(0, configuration.number_of_steps) * step_size
                )

            current_time = new_time

        return _drop_data_before(
            pa.Table.from_pydict({"ts": ts, "value": value}), start_date
        )

    def list_series(self, selector: SeriesSearch) -> Generator[Metadata, None, None]:
        """Yields all possible metadata combinations using the signal configuration and the provided selector."""
        arg_list = []
        if self.__config is None:
            return

        rng = Random(self.__config.initial_seed)
        arg_list.append(
            _extract_from_tag(
                selector.tags,
                "seed",
                [
                    rng.randint(0, sys.maxsize)
                    for _ in range(self.__config.number_of_seeds)
                ],
            )
        )
        arg_list.append(
            _extract_from_tag(
                selector.tags,
                "interval_seconds_min",
                self.__config.interval_seconds_min,
            )
        )
        arg_list.append(
            _extract_from_tag(
                selector.tags,
                "interval_seconds_max",
                self.__config.interval_seconds_max,
            )
        )
        arg_list.append(
            [
                dict(min=value.min, max=value.max, number_of_steps=number_of_steps)
                for value in self.__config.values
                for number_of_steps in value.number_of_steps
            ]
        )

        for entry in itertools.product(*arg_list):
            yield _build_step_search_result(self.__config, entry, selector.source)


def _ensure_list(value) -> List:
    if isinstance(value, List):
        return value
    return [value]


def _build_step_search_result(
    config: StepSignalGeneratorConfig, entry: tuple, source_name: str
) -> Metadata:
    series_selector = SeriesSelector(
        source_name,
        {
            "series name": config.series_name,
            "signal_type": config.signal_type,
            "seed": str(entry[0]),
            "interval_seconds_min": str(entry[1]),
            "interval_seconds_max": str(entry[2]),
            "min_value": str(entry[3]["min"]),
            "max_value": str(entry[3]["max"]),
            "number_of_steps": str(entry[3]["number_of_steps"]),
        },
    )
    metadata = Metadata(series_selector)
    for field_name, field_value in config.metadata.items():
        metadata.coerce_field(field_name, field_value)
    return metadata


def _get_step_configuration(selector: SeriesSelector) -> StepSignalConfig:
    return StepSignalConfig(
        selector.tags["series name"],
        selector.tags["signal_type"],
        int(selector.tags["seed"]),
        int(selector.tags["interval_seconds_min"]),
        int(selector.tags["interval_seconds_max"]),
        float(selector.tags["min_value"]),
        float(selector.tags["max_value"]),
        int(selector.tags["number_of_steps"]),
    )


def _get_start_of_day(ts: datetime) -> datetime:
    current_day = ts.date()
    return datetime(
        current_day.year, current_day.month, current_day.day, tzinfo=timezone.utc
    )


class WhiteNoiseSignalGenerator:
    """White noise signal generator."""

    __config: Optional[WhiteNoiseSignalGeneratorConfig] = None

    def __init__(self, config: Optional[Dict] = None):
        if config is not None:
            self.__config = WhiteNoiseSignalGeneratorConfig(
                config["seriesName"],
                config["type"],
                config.get("initialSeed", 0),
                config.get("numberOfSeeds", 1),
                config["samplingInterval"]["intervalSecondsMin"],
                config["samplingInterval"]["intervalSecondsMax"],
                config.get("metadata", {}),
                config["values"]["mean"],
                config["values"]["standardDeviation"],
            )

    def generate(  # pylint: disable=no-self-use
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Generates white noise based on a selector start and end date."""
        current_time = _get_start_of_day(start_date)
        configuration = _get_white_noise_configuration(selector)
        ts = []
        value = []

        rng = Random(_get_hex_digest(current_time, configuration.to_bytes()))

        numpy_rng = numpy.random.default_rng(
            _get_int_digest(current_time, configuration.to_bytes())
        )

        while current_time <= end_date:
            generated_value = numpy_rng.normal(
                configuration.mean, configuration.standard_deviation, 1
            )[0]
            value.append(generated_value)
            ts.append(current_time)

            time_increment = rng.randint(
                configuration.interval_seconds_min, configuration.interval_seconds_max
            )
            new_time = current_time + timedelta(seconds=time_increment)

            if new_time.date() != current_time.date():
                new_time = _get_start_of_day(new_time)
                rng.seed(_get_hex_digest(new_time, configuration.to_bytes()))
                numpy_rng = numpy.random.default_rng(
                    _get_int_digest(new_time, configuration.to_bytes())
                )

            current_time = new_time

        return _drop_data_before(
            pa.Table.from_pydict({"ts": ts, "value": value}), start_date
        )

    def list_series(self, selector: SeriesSearch) -> Generator[Metadata, None, None]:
        """Yields all possible metadata combinations using the signal configuration and the provided selector."""
        arg_list = []
        if self.__config is None:
            return

        rng = Random(self.__config.initial_seed)
        arg_list.append(
            _extract_from_tag(
                selector.tags,
                "seed",
                [
                    rng.randint(0, sys.maxsize)
                    for _ in range(self.__config.number_of_seeds)
                ],
            )
        )
        arg_list.append(
            _extract_from_tag(
                selector.tags,
                "interval_seconds_min",
                self.__config.interval_seconds_min,
            )
        )
        arg_list.append(
            _extract_from_tag(
                selector.tags,
                "interval_seconds_max",
                self.__config.interval_seconds_max,
            )
        )
        arg_list.append(_extract_from_tag(selector.tags, "mean", self.__config.mean))
        arg_list.append(
            _extract_from_tag(
                selector.tags,
                "standard_deviation",
                self.__config.standard_deviation,
            )
        )

        for entry in itertools.product(*arg_list):
            yield _build_white_noise_search_result(
                self.__config, entry, selector.source
            )


def _build_white_noise_search_result(
    config: WhiteNoiseSignalGeneratorConfig, entry: tuple, source_name: str
) -> Metadata:
    series_selector = SeriesSelector(
        source_name,
        {
            "series name": config.series_name,
            "signal_type": config.signal_type,
            "seed": str(entry[0]),
            "interval_seconds_min": str(entry[1]),
            "interval_seconds_max": str(entry[2]),
            "mean": str(entry[3]),
            "standard_deviation": str(entry[4]),
        },
    )
    metadata = Metadata(series_selector)
    for field_name, field_value in config.metadata.items():
        metadata.coerce_field(field_name, field_value)
    return metadata


def _get_white_noise_configuration(
    selector: SeriesSelector,
) -> WhiteNoiseSignalConfig:
    return WhiteNoiseSignalConfig(
        selector.tags["series name"],
        selector.tags["signal_type"],
        int(selector.tags["seed"]),
        int(selector.tags["interval_seconds_min"]),
        int(selector.tags["interval_seconds_max"]),
        float(selector.tags["mean"]),
        float(selector.tags["standard_deviation"]),
    )


class SineSignalGenerator:
    """Sine signal generator."""

    __config: Optional[SineSignalGeneratorConfig] = None

    def __init__(self, config: Optional[Dict] = None):
        if config is not None:
            self.__config = SineSignalGeneratorConfig(
                config["seriesName"],
                config["type"],
                config.get("initialSeed", 0),
                config.get("numberOfSeeds", 1),
                config["samplingInterval"]["intervalSecondsMin"],
                config["samplingInterval"]["intervalSecondsMax"],
                config.get("metadata", {}),
                config["values"]["periodSeconds"],
                config["values"]["phaseSeconds"],
                config["values"]["amplitude"],
                config["values"]["shift"],
            )

    def generate(  # pylint: disable=no-self-use
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Generates a sine function based on a selector start and end date."""
        current_time = _get_start_of_day(start_date)
        configuration = _get_sine_configuration(selector)
        ts = []
        value = []

        rng = Random(_get_hex_digest(current_time, configuration.to_bytes()))

        while current_time <= end_date:
            value.append(
                calculate_sine(
                    current_time,
                    period_seconds=configuration.period_seconds,
                    phase_seconds=configuration.phase_seconds,
                    amplitude=configuration.amplitude,
                    shift=configuration.shift,
                )
            )
            ts.append(current_time)

            time_increment = rng.randint(
                configuration.interval_seconds_min, configuration.interval_seconds_max
            )
            new_time = current_time + timedelta(seconds=time_increment)

            if new_time.date() != current_time.date():
                new_time = _get_start_of_day(new_time)

                rng.seed(_get_hex_digest(new_time, configuration.to_bytes()))

            current_time = new_time

        return _drop_data_before(
            pa.Table.from_pydict({"ts": ts, "value": value}), start_date
        )

    def list_series(self, selector: SeriesSearch) -> Generator[Metadata, None, None]:
        """Yields all possible metadata combinations using the signal configuration and the provided selector."""
        if self.__config is None:
            return

        arg_list = []
        rng = Random(self.__config.initial_seed)
        arg_list.append(
            _extract_from_tag(
                selector.tags,
                "seed",
                [
                    rng.randint(0, sys.maxsize)
                    for _ in range(self.__config.number_of_seeds)
                ],
            )
        )
        arg_list.append(
            _extract_from_tag(
                selector.tags,
                "interval_seconds_min",
                self.__config.interval_seconds_min,
            )
        )
        arg_list.append(
            _extract_from_tag(
                selector.tags,
                "interval_seconds_max",
                self.__config.interval_seconds_max,
            )
        )
        arg_list.append(
            _extract_from_tag(
                selector.tags, "period_seconds", self.__config.period_seconds
            )
        )
        arg_list.append(
            _extract_from_tag(
                selector.tags,
                "amplitude",
                self.__config.amplitude,
            )
        )
        arg_list.append(
            _extract_from_tag(
                selector.tags, "phase_seconds", self.__config.phase_seconds
            )
        )
        arg_list.append(_extract_from_tag(selector.tags, "shift", self.__config.shift))

        for entry in itertools.product(*arg_list):
            yield _build_sine_search_result(self.__config, entry, selector.source)


def _build_sine_search_result(
    config: SineSignalGeneratorConfig, entry: tuple, source_name: str
) -> Metadata:
    series_selector = SeriesSelector(
        source_name,
        {
            "series name": config.series_name,
            "signal_type": config.signal_type,
            "seed": str(entry[0]),
            "interval_seconds_min": str(entry[1]),
            "interval_seconds_max": str(entry[2]),
            "period_seconds": str(entry[3]),
            "amplitude": str(entry[4]),
            "phase_seconds": str(entry[5]),
            "shift": str(entry[6]),
        },
    )
    metadata = Metadata(series_selector)
    for field_name, field_value in config.metadata.items():
        metadata.coerce_field(field_name, field_value)
    return metadata


def _get_sine_configuration(selector: SeriesSelector) -> SineSignalConfig:
    return SineSignalConfig(
        selector.tags["series name"],
        selector.tags["signal_type"],
        int(selector.tags["seed"]),
        int(selector.tags["interval_seconds_min"]),
        int(selector.tags["interval_seconds_max"]),
        float(selector.tags["period_seconds"]),
        float(selector.tags["phase_seconds"]),
        float(selector.tags["amplitude"]),
        float(selector.tags["shift"]),
    )


def calculate_sine(
    ts: datetime,
    *,
    period_seconds: float,
    phase_seconds: float = 0,
    amplitude: float = 1,
    shift: float = 0,
) -> float:
    """Calculate the sine function at unix timestamp ts."""
    return (
        amplitude
        * numpy.sin(2 * numpy.pi * (ts.timestamp() + phase_seconds) / period_seconds)
        + shift
    )


class SimulatorSource:
    """A simulator data source."""

    __signal_generators: Dict[str, List[SignalGenerator]]

    __yaml_path: Optional[Path] = None

    def __init__(self, config: SimulatorConfiguration):
        self.__signal_generators = defaultdict(list)

        if config.path is not None:
            if not HAS_YAML:
                raise MissingModuleException("PyYAML")
            self.__yaml_path = Path(config.path)

        if self.__yaml_path is None:
            self.__signal_generators["step"].append(StepSignalGenerator())
            self.__signal_generators["white noise"].append(WhiteNoiseSignalGenerator())
            self.__signal_generators["sine"].append(SineSignalGenerator())
        else:
            with self.__yaml_path.open(encoding="utf-8") as file:
                yaml_data = yaml.safe_load(file)
                for signal_config in yaml_data.get("signals", []):
                    self._register_generator(signal_config)

    def search(
        self, selector: SeriesSearch
    ) -> Generator[Union[Metadata, SeriesSelector], None, None]:
        """Yields all possible metadata combinations using the signal configuration and the provided selector."""
        all_series = []
        if "signal_type" in selector.tags:
            for generator in self.__signal_generators[selector.tags["signal_type"]]:
                all_series.append(generator.list_series(selector))
        else:
            for generators in self.__signal_generators.values():
                for generator in generators:
                    all_series.append(generator.list_series(selector))

        for series in all_series:
            yield from series

    # pylint: disable=no-self-use
    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Data explorer currently always returns empty metadata."""
        return Metadata(selector)

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Generates data in steps based on a selector start and end date."""
        if "signal_type" not in selector.tags:
            return pa.Table.from_pydict({"ts": [], "value": []})
        for generator in self.__signal_generators[selector.tags["signal_type"]]:
            return generator.generate(selector, start_date, end_date)

    def get_source_structure(self, _: SeriesSelector) -> Optional[SourceStructure]:
        """Return the structure of a source."""

    def _register_generator(self, config: Dict):
        """Register a generator."""
        signal_type = config["type"]
        if signal_type == "step":
            self.__signal_generators[signal_type].append(StepSignalGenerator(config))
        if signal_type == "white noise":
            self.__signal_generators[signal_type].append(
                WhiteNoiseSignalGenerator(config)
            )
        if signal_type == "sine":
            self.__signal_generators[signal_type].append(SineSignalGenerator(config))


def _extract_from_tag(
    tags: Dict[str, str], key: str, fallback: Union[List[Any], Any]
) -> List[Any]:
    if key in tags:
        return [tags[key]]
    if not isinstance(fallback, List):
        return [fallback]
    return fallback


def _drop_data_before(table: pa.Table, ts: datetime) -> pa.Table:
    # pylint: disable=no-member
    keep_after = pa.compute.greater_equal(table["ts"], pa.scalar(ts))
    return table.filter(keep_after)


def _get_hex_digest(ts: datetime, extra_bytes: bytes) -> str:
    return _get_hash(ts, extra_bytes).hexdigest()


def _get_int_digest(ts: datetime, extra_bytes) -> int:
    return int.from_bytes(_get_hash(ts, extra_bytes).digest()[:4], "little")


def _get_hash(ts: datetime, extra_bytes: bytes):
    return sha1(extra_bytes + bytes(ts.isoformat(), "UTF-8"))
